###########################################################
# Helper functions used during the CAS webinar

CopyClip <- function(df, col.names=TRUE, row.names=FALSE) {
    # Copy a data frame onto the clipboard
    #
    # If col.names or row.names is TRUE, include those as a row/column
    # Could just use write.table(df, file="clipboard-128", sep="\t")
    # but then would be limited to 32kB.
    filename <- tempfile()
    write.table(df, file=filename, sep="\t",
                row.names=row.names, col.names=col.names)
    lines <- readLines(filename)
    writeClipboard(paste(lines, collapse="\n"))
    unlink(filename)
}

PasteClip <- function(header=TRUE) {
    # Copy a data frame from the clipboard
    #
    # The header argument indicates whether the first row contains field names
    read.table(file="clipboard", sep="\t", header=header, quote="",
               na.strings=c("#VALUE!", "#DIV/0!", "#NAME?", "#NUM!", "#N/A",
                   "#NULL!", "#REF!"),
               stringsAsFactors=FALSE, comment.char="")
}

LiftCurve <- function(comp.df, mod1, mod2, file.post=NULL,
                      by.exposure=TRUE, variance.p=1.5) {
    # Return a lift curve comparing the two columns.
    # Save to disk if file.post is set
    PrepareComp <- function() {
        # Return an augmented version of the comp.df that has ratios and deciles
        # Also do reasonability checks on comp.df
        stopifnot(all(c(mod1, mod2) %in% names(comp.df)))
        comp.df$m1.fit <- comp.df[[mod1]]
        comp.df$m2.fit <- comp.df[[mod2]]
        stopifnot(with(comp.df, all((m1.fit > 1e-08 & m2.fit > 1e-08)
                                    | (m1.fit < 1e-05 & m2.fit < 1e-05))))
        stopifnot(with(comp.df, all(actual == 0 | m1.fit > 1e-08)))

        comp.df <- subset(comp.df, m1.fit > 1e-05)
        comp.df$mratio <- comp.df[[mod1]]/comp.df[[mod2]]
        comp.df <- comp.df[order(comp.df$mratio), ]
        comp.df$cum.exp <- cumsum(comp.df$exposure)
        breaks <- quantile(comp.df$cum.exp,
                           seq(from = 0, to = 1, by = 0.1))
        comp.df$decile <- as.numeric(cut(comp.df$cum.exp, breaks = breaks,
            include.lowest = TRUE))
        return(comp.df)
    }

    VarianceConstants <- function(comp.df) {
        # Return the variance constants as c(m1.varconst, m2.varconst)
        # the variance for each record = (var constant) /  exposures * mean^1.5
        #
        # 1.5 is selected to be a typical tweedie p value, and an
        # average of the poisson and gamma distributions.
        #
        # The variance constant is based on a normalish distribution and
        # maximum likelihood.
        m1.const <- with(comp.df,
                         mean((actual - m1.fit)^2 * exposure/m1.fit^1.5))
        m2.const <- with(comp.df,
                         mean((actual - m2.fit)^2 * exposure/m2.fit^1.5))
        return(c(m1.const, m2.const))
    }

    MakePlotDF <- function(comp.df, var.consts) {
        # Return the plotting data frame
        summary.df <- sqldf(paste(
                    "select decile, sum([m1_fit]) as [m1.fit],
                     sum(power([m1_fit], 1.5) /
                         case when exposure = 0 then 0 else exposure end)
                       as [m1.var],
                     sum([m2_fit]) as [m2.fit],
                     sum(power([m2_fit], 1.5) /
                          case when exposure = 0 then 0 else exposure end)
                       as [m2.var],
                     sum(actual) as actual
                     from [comp.df]
                     group by decile order by decile", sep=""))
        summary.df <- transform(summary.df,
                                m1.pct=(m1.fit - actual) / actual,
                                m2.pct=(m2.fit - actual) / actual,
                                m1.pctsd=sqrt(m1.var * var.consts[1]) / actual,
                                m2.pctsd=sqrt(m2.var * var.consts[2]) / actual)
        plot.df <- with(summary.df,
                        data.frame(decile=rep(decile, 2),
                                   Model=c(rep(mod1, nrow(summary.df)),
                                       rep(mod2, nrow(summary.df))),
                                   pct=c(m1.pct, m2.pct),
                                   pctsd=c(m1.pctsd, m2.pctsd)))
        return(plot.df)
    }

    MakePlot <- function(plot.df) {
        # Return the actual plot object
        title <- paste("Double Lift Chart: ", mod1, " vs ", mod2, sep="")
        if (!is.null(file.post))
            title <- paste(title, ", ", file.post, sep="")

        plot <- (ggplot(plot.df)
                 + geom_smooth(aes(x=decile, y=100 * pct,
                                   ymin=100 * (pct - pctsd),
                                   ymax=100 * (pct + pctsd),
                                   group=Model, color=Model),
                               stat="identity", size=1, fill="grey80")
                 + labs(x="Decile of Ratios of Fitted Losses",
                        y="(Predicted - Actual) / Actual (%)",
                        title=title)
                 + geom_hline(aes(yintercept=0),
                              color="black", size=0.3, linetype="dotted"))
        return(plot)
    }

    comp.df <- PrepareComp()
    var.consts <- VarianceConstants(comp.df)
    plot <- MakePlot(MakePlotDF(comp.df, var.consts))
    if (!is.null(file.post)) {
        filename <- paste(mod1, "-", mod2, "-", file.post, ".png", sep="")
        ggsave(plot=plot, filename=filename, width=8, height=5.5)
    }
    return(plot)
}

`%StartsWith%` <- function(charvec, prefix) {
    # Return true if each char in charvec starts with prefix
    return(substr(charvec, 1, nchar(prefix)) == prefix)
}
`%EndsWith%` <- function(charvec, suffix) {
    # Return true if each char in charvec ends with suffix
    len <- nchar(charvec)
    return(substr(charvec, len - nchar(suffix) + 1, len) == suffix)
}
`%Contains%` <- function(charvec, substr) {
    # Return true if each char in charvec contains substr
    result <- rep(FALSE, length(charvec))
    result[grep(substr, charvec, fixed=TRUE)] <- TRUE
    return(result)
}
`%s%` <- function(a, b) paste(a, b, sep="")
