############################################################
# Fit the data

library(tweedie)
library(splines)
library(ggplot2)
library(sqldf)
library(reshape)

# Data import
total.df <- read.csv(file="input.data.csv")
train.df <- total.df[1:50e3, ]
test.df <- total.df[(50e3 + 1):75e3, ]

# Find tweedie P value
set.seed(0)
x <- tweedie.profile(loss ~ 1, offset=log(exposure),
                     do.plot=TRUE, do.smooth=FALSE, method="series",
                     data=train.df[sample(1:nrow(train.df), 1e3), ])
tweedie.p <- x$xi.max
rm(x)

# Initial Fit
glm1 <- glm(loss ~ sex + marital + age + num_drv + num_veh
                   + credit + major + minor,
            family=tweedie(var.power=tweedie.p, link.power=0),
            data=train.df, offset=log(exposure),
            weights=exposure^(tweedie.p - 1))
summary(glm1)

# Try adding a spline for age
glm2 <- glm(loss ~ sex + marital + ns(age, knots=c(30, 65))
            + num_drv + num_veh + credit + major + minor,
            family=tweedie(var.power=tweedie.p, link.power=0),
            data=train.df, offset=log(exposure),
            weights=exposure^(tweedie.p - 1))
summary(glm2)

x <- ns(15:100, knots=c(30, 65))
x.df <- melt(data.frame(x, age=15:100), id.vars="age")
qplot(data=x.df, x=age, y=value, group=variable, color=variable, geom="line")

agecurve <- x %*% coef(glm2)[names(coef(glm2)) %StartsWith% "ns"]
qplot(x=15:100, y=agecurve, geom="line")

# More splines
glm3 <- glm(loss ~ (as.factor(sex) * as.factor(marital)
                    * ns(age, knots=c(20, 25, 30, 40, 50, 70)))
            + as.factor(num_drv) * as.factor(num_veh)
            + ns(credit, knots=seq(from=300, to=700, by=100))
            + major + minor,
            family=tweedie(var.power=tweedie.p, link.power=0),
            data=train.df, offset=log(exposure),
            weights=exposure^(tweedie.p - 1))
summary(glm3)
CopyClip(summary(glm3)$coefficients, row.names=TRUE)

# Cut back a little bit
glm4 <- glm(loss ~ sex
            + as.factor(marital) * ns(age, knots=c(20, 30, 50, 70))
            + as.factor(num_drv) + as.factor(num_veh)
            + ns(credit, knots=c(400, 600))
            + major + minor,
            family=tweedie(var.power=tweedie.p, link.power=0),
            data=train.df, offset=log(exposure),
            weights=exposure^(tweedie.p - 1))
summary(glm4)


################################## Lift curves
# on training data
comp.df <- data.frame(actual=train.df$loss, exposure=train.df$exposure,
                      glm1=predict(glm1, newdata=train.df, type="response"),
                      glm2=predict(glm2, newdata=train.df, type="response"),
                      glm3=predict(glm3, newdata=train.df, type="response"),
                      glm4=predict(glm4, newdata=train.df, type="response"))

comp.df$ratio <- with(comp.df, glm1 / glm2)
ratio.deciles <- quantile(comp.df$ratio, probs=0:10/10, include.lowest=TRUE)
comp.df$decile <- as.numeric(cut(comp.df$ratio, breaks=ratio.deciles,
                                 include.lowest=TRUE))
plot.df <- sqldf("select decile,
                      sum(glm1) / sum(actual) as glm1,
                      sum(glm2) / sum(actual) as glm2
                  from [comp.df]
                  group by decile order by decile")
(ggplot(melt(plot.df, id.vars="decile"))
 + geom_line(aes(x=decile, y=value, group=variable, color=variable)))

LiftCurve(comp.df, "glm1", "glm2")
LiftCurve(comp.df, "glm2", "glm3")
LiftCurve(comp.df, "glm3", "glm4") # bigger model always fits better

# on test data
comp.test.df <- data.frame(actual=test.df$loss, exposure=test.df$exposure,
                      glm1=predict(glm1, newdata=test.df, type="response"),
                      glm2=predict(glm2, newdata=test.df, type="response"),
                      glm3=predict(glm3, newdata=test.df, type="response"),
                      glm4=predict(glm4, newdata=test.df, type="response"))
LiftCurve(comp.test.df, "glm1", "glm2")
LiftCurve(comp.test.df, "glm2", "glm3")
LiftCurve(comp.test.df, "glm3", "glm4")
LiftCurve(comp.test.df, "glm2", "glm4")


################################## Query out results
sel.coef <- coef(glm4)

# Credit curve
cred.curve <- (ns(200:800, knots=c(400, 600))
               %*% (sel.coef[names(sel.coef) %Contains% "credit"]))
credit.table <- data.frame(credit.score=200:800, factor=exp(cred.curve))
credit.table <- transform(credit.table,
                          factor=factor / factor[credit.score == 500])
qplot(x=credit.score, y=factor, data=credit.table, geom="line")

# age/sex/marital status curves
age.names <- (names(sel.coef) %Contains% "age"
              & !(names(sel.coef) %Contains% "marital"))
age.names.mar <- (names(sel.coef) %Contains% "age"
                  & names(sel.coef) %Contains% "marital")
asm.df <- data.frame(age=15:100,
                     married.female=exp(ns(15:100, knots=c(20, 30, 50, 70))
                         %*% sel.coef[age.names]),
                     unmarried.female=exp(ns(15:100, knots=c(20, 30, 50, 70))
                         %*% (sel.coef[age.names] +
                              sel.coef[age.names.mar])))
asm.df <- transform(asm.df,
                    married.male=married.female * exp(sel.coef["sex"]),
                    unmarried.male=unmarried.female * exp(sel.coef["sex"]))
melted.asm.df <- melt(asm.df, id.vars="age")
melted.asm.df <- transform(melted.asm.df,
                           value=value / value[age == 30
                               & variable == "married.male"])
(ggplot(melted.asm.df)
 + geom_line(aes(x=age, y=value, group=variable, color=variable)))
(ggplot(melted.asm.df)
 + geom_line(aes(x=age, y=value))
 + facet_grid(variable ~ .))

###### Compare to true results
# Credit
credit.table$truth <- 1 + ((credit.table$credit.score - 800) / 100)^2 / 10
credit.table$truth <- with(credit.table,
                           truth / truth[credit.score == 500])
(ggplot(melt(credit.table, id.vars="credit.score"))
 + geom_line(aes(x=credit.score, y=value, group=variable, color=variable)))

# ASM
asm.comp.df <- transform(melted.asm.df,
                         rel_unmar=ifelse(age <= 70,
                             1 + ((age - 50) / 20)^2,
                             2 + 0.02 * (age - 70)),
                         rel_mar=ifelse(age <= 70,
                             1 + ((age - 55) / 30)^2,
                             1.25 * 1.02^(age - 70)),
                         rel_sex=1 + .1 * !(variable %Contains% "female"))
asm.comp.df <- transform(asm.comp.df,
                   truth=rel_sex * ifelse(variable %Contains% "unmarried",
                               rel_unmar, rel_mar))
asm.comp.df <- with(asm.comp.df,
                      data.frame(age=age, sm=as.character(variable),
                                 factor=value,
                                 truth=truth / truth[age == 30 &
                                     variable == "married.male"]))
(ggplot(melt(asm.comp.df, id.vars=c("age", "sm")))
 + geom_line(aes(x=age, y=value, group=variable, color=variable))
 + facet_grid(sm ~ .))
(ggplot(melt(asm.comp.df, id.vars=c("age", "sm")))
 + geom_line(aes(x=age, y=value, group=sm, color=sm))
 + facet_grid(variable ~ .))
